@@ -10,8 +10,10 @@ from .decorators import (  # noqa: F401
     pyplot,
     immediate,
     changes_per_run,
+    accepts,
+    returns,
 )
 
 from . import protocol  # noqa: F401
 
-__version__ = u"0.7.0"
+__version__ = u"0.8.0"
