@@ -2,46 +2,70 @@
 Contains various data structures used by Bionic's infrastructure.
 """
 
-from collections import namedtuple
+import attr
 
 from .util import ImmutableSequence, ImmutableMapping
 
 
-# TODO Consider using the attr library here?
-class TaskKey(namedtuple("TaskKey", "dnode case_key")):
+@attr.s(frozen=True)
+class EntityDefinition:
+    """
+    Describes the immutable properties of an entity. These properties generally have
+    to do with the entity's "contract": the assumptions other parts of the system can
+    make about its value. However, this does *not* include the way the entity's value
+    is determined; this is configured separately and can be changed more easily.
+    """
+
+    name = attr.ib()
+    protocol = attr.ib()
+    doc = attr.ib()
+    can_persist = attr.ib()
+    optional_should_memoize = attr.ib()
+
+
+@attr.s(frozen=True)
+class TaskKey:
     """
     A unique identifier for a Task.
     """
 
-    def __new__(cls, dnode, case_key):
-        return super(TaskKey, cls).__new__(cls, dnode, case_key)
-
-    def __repr__(self):
-        return f"TaskKey({self.dnode!r}, {self.case_key!r})"
+    dnode = attr.ib()
+    case_key = attr.ib()
 
     def __str__(self):
         args_str = ", ".join(f"{name}={value}" for name, value in self.case_key.items())
-        return f"{self.dnode.to_entity_name()}({args_str})"
+        return f"{self.dnode.to_descriptor(near_commas=True)}({args_str})"
 
 
-class Task(object):
+@attr.s(frozen=True)
+class Task:
     """
     A unit of work.  Can have dependencies, which are referred to via their
     TaskKeys.
+
+    Attributes
+    ----------
+    keys: list of TaskKeys
+        Keys corresponding to the output values computed by this task.
+    dep_keys: list of TaskKeys
+        Keys corresponding to the input values required by this task.
+    compute_func: function taking a single ``dep_values`` argument
+        Generates output values based on the passed input values.
+    is_simple_lookup: boolean
+        Whether this task consists of simply looking up the fixed value of an entity;
+        used to determine what message to log when this task is computed.
     """
 
-    def __init__(self, keys, dep_keys, compute_func, is_simple_lookup=False):
-        self.keys = tuple(keys)
-        self.dep_keys = tuple(dep_keys)
-        self.compute = compute_func
-        self.is_simple_lookup = is_simple_lookup
+    keys = attr.ib(converter=tuple)
+    dep_keys = attr.ib(converter=tuple)
+    compute_func = attr.ib()
+    is_simple_lookup = attr.ib(default=False)
 
-    def key_for_entity_name(self, name):
-        matching_keys = [
-            task_key
-            for task_key in self.keys
-            if task_key.dnode.to_entity_name() == name
-        ]
+    def compute(self, dep_values):
+        return self.compute_func(dep_values)
+
+    def key_for_dnode(self, dnode):
+        matching_keys = [task_key for task_key in self.keys if task_key.dnode == dnode]
         (key,) = matching_keys
         return key
 
@@ -49,48 +73,54 @@ class Task(object):
         return f"Task({self.keys!r}, {self.dep_keys!r})"
 
 
-class Query(object):
+@attr.s(frozen=True)
+class Query:
     """
     Represents a request for a specific entity value.
     """
 
-    def __init__(self, task_key, protocol, provenance):
-        self.task_key = task_key
-        self.dnode = task_key.dnode
-        self.case_key = task_key.case_key
-        self.protocol = protocol
-        self.provenance = provenance
+    task_key = attr.ib()
+    protocol = attr.ib()
+    provenance = attr.ib()
+
+    @property
+    def dnode(self):
+        return self.task_key.dnode
+
+    @property
+    def case_key(self):
+        return self.task_key.case_key
 
     def __repr__(self):
         return f"Query({self.task_key}, {self.provenance!r})"
 
 
-class Result(object):
+@attr.s(frozen=True)
+class Result:
     """
     Represents one value for one entity.
     """
 
-    def __init__(self, query, value, value_hash=None, file_path=None):
-        self.query = query
-        self.value = value
-        # Only present when value should persist.
-        self.file_path = file_path
-        self.value_hash = value_hash
+    query = attr.ib()
+    value = attr.ib()
+    file_path = attr.ib(default=None)
+    value_hash = attr.ib(default=None)
+    value_is_missing = attr.ib(default=False)
 
     def __repr__(self):
         return f"Result({self.query!r}, {self.value!r})"
 
 
-class ProvenanceDigest(object):
+@attr.s(frozen=True)
+class ProvenanceDigest:
     """
     A collection of values used by Provenance for different chained hashes.
     These hashes depend on the entities and can come from either another
     provenance or the value hash of a result.
     """
 
-    def __init__(self, functional_hash, exact_hash):
-        self.functional_hash = functional_hash
-        self.exact_hash = exact_hash
+    functional_hash = attr.ib()
+    exact_hash = attr.ib()
 
     @classmethod
     def from_provenance(cls, provenance):
@@ -151,28 +181,32 @@ class CaseKeySpace(ImmutableSequence):
 
 class CaseKey(ImmutableMapping):
     """
-    A collection of name-value pairs that uniquely identifies a case.
+    A collection of name-token pairs that uniquely identifies a case.
     """
 
-    def __init__(self, name_value_token_triples):
-        values_by_name = {
-            name: value for name, value, token in name_value_token_triples
-        }
-        tokens_by_name = {
-            name: token for name, value, token in name_value_token_triples
-        }
+    def __init__(self, name_token_pairs):
+        tokens_by_name = {name: token for name, token in name_token_pairs}
 
         super(CaseKey, self).__init__(tokens_by_name)
-        self._nvt_triples = name_value_token_triples
-        self.values = values_by_name
+        self._name_token_pairs = name_token_pairs
         self.tokens = tokens_by_name
-        self.space = CaseKeySpace(list(values_by_name.keys()))
+        self.space = CaseKeySpace(list(tokens_by_name.keys()))
+        self.missing_names = [
+            # None is a sentinel value used to indicate that no value is available.
+            # Normally I would prefer to represent missing-ness out-of-band by making the
+            # `missing_names` field the source of truth here, but the relational methods like
+            # `project` are cleaner when we use a sentinel value.
+            name
+            for name, token in name_token_pairs
+            if token is None
+        ]
+        self.has_missing_values = len(self.missing_names) > 0
 
     def project(self, key_space):
         return CaseKey(
             [
-                (name, value, token)
-                for name, value, token in self._nvt_triples
+                (name, token)
+                for name, token in self._name_token_pairs
                 if name in key_space
             ]
         )
@@ -180,26 +214,22 @@ class CaseKey(ImmutableMapping):
     def drop(self, key_space):
         return CaseKey(
             [
-                (name, value, token)
-                for name, value, token in self._nvt_triples
+                (name, token)
+                for name, token in self._name_token_pairs
                 if name not in key_space
             ]
         )
 
     def merge(self, other):
-        vt_pairs_by_name = {
-            name: (value, token) for name, value, token in self._nvt_triples
-        }
+        tokens_by_name = {name: token for name, token in self._name_token_pairs}
 
-        for name, value, token in other._nvt_triples:
-            if name in vt_pairs_by_name:
-                assert token == vt_pairs_by_name[name][1]
+        for name, token in other._name_token_pairs:
+            if name in tokens_by_name:
+                assert token == tokens_by_name[name]
             else:
-                vt_pairs_by_name[name] = value, token
+                tokens_by_name[name] = token
 
-        return CaseKey(
-            [(name, value, token) for name, (value, token) in vt_pairs_by_name.items()]
-        )
+        return CaseKey([(name, token) for name, token in tokens_by_name.items()])
 
     def __repr__(self):
         args_str = ", ".join(f"{name}={token}" for name, token in self.items())
@@ -221,22 +251,6 @@ class ResultGroup(ImmutableSequence):
         return f"ResultGroup({list(self)!r})"
 
 
-class CodeVersion(object):
-    """
-    Contains the user-designated version of a piece of code, consisting of a
-    major and a minor version string.  The convention is that changing the
-    major version indicates a functional change, while changing the minor
-    version indicates a nonfunctional change.
-    """
-
-    def __init__(self, major, minor):
-        self.major = str_from_version_value(major)
-        self.minor = str_from_version_value(minor)
-
-    def __repr__(self):
-        return f"CodeVersion({self.major!r}, {self.minor!r})"
-
-
 def str_from_version_value(value):
     if value is None:
         return "0"
@@ -248,11 +262,35 @@ def str_from_version_value(value):
         raise ValueError(f"Version values must be str, int, or None: got {value!r}")
 
 
-# A collection of characteristics attempting to uniquely identify a function.
-CodeFingerprint = namedtuple("CodeFingerprint", "version bytecode_hash orig_flow_name")
+@attr.s(frozen=True)
+class CodeVersion:
+    """
+    Contains the user-designated version of a piece of code, consisting of a
+    major and a minor version string.  The convention is that changing the
+    major version indicates a functional change, while changing the minor
+    version indicates a nonfunctional change.
+    """
+
+    major = attr.ib(converter=str_from_version_value)
+    minor = attr.ib(converter=str_from_version_value)
 
 
-# Encodes the versioning rules to use when computing entity values.
-VersioningPolicy = namedtuple(
-    "VersioningPolicy", "check_for_bytecode_errors treat_bytecode_as_functional"
-)
+@attr.s(frozen=True)
+class CodeFingerprint:
+    """
+    A collection of characteristics attempting to uniquely identify a function.
+    """
+
+    version = attr.ib()
+    bytecode_hash = attr.ib()
+    orig_flow_name = attr.ib()
+
+
+@attr.s(frozen=True)
+class VersioningPolicy:
+    """
+    Encodes the versioning rules to use when computing entity values.
+    """
+
+    check_for_bytecode_errors = attr.ib()
+    treat_bytecode_as_functional = attr.ib()
