@@ -1,7 +1,7 @@
 """
 These are the decorators we expose to Bionic users.  They are used as follows:
 
-    @builder.derive
+    @builder
     @bionic.decorator1
     @bionic.decorator2
     ...
@@ -11,13 +11,16 @@ These are the decorators we expose to Bionic users.  They are used as follows:
 """
 
 from .datatypes import CodeVersion
+from .decoration import decorator_updating_accumulator
+from .descriptors.parsing import dnode_from_descriptor, entity_dnode_from_descriptor
 from .provider import (
     GatherProvider,
     AttrUpdateProvider,
     PyplotProvider,
     RenamingProvider,
     NameSplittingProvider,
-    provider_wrapper,
+    ArgDescriptorSubstitutionProvider,
+    NewOutputDescriptorProvider,
 )
 from . import interpret
 
@@ -59,8 +62,10 @@ def version(major=None, minor=None):
         A decorator which can be applied to an entity function.
     """
 
-    return provider_wrapper(
-        AttrUpdateProvider, "code_version", CodeVersion(major, minor)
+    return decorator_updating_accumulator(
+        lambda acc: acc.wrap_provider(
+            AttrUpdateProvider, "code_version", CodeVersion(major, minor)
+        )
     )
 
 
@@ -87,12 +92,15 @@ def persist(enabled):
     if not isinstance(enabled, bool):
         raise ValueError(f"Argument must be a boolean; got {enabled!r}")
 
-    return provider_wrapper(AttrUpdateProvider, "_can_persist", enabled)
+    return decorator_updating_accumulator(
+        lambda acc: acc.update_attr("can_persist", enabled, "@persist")
+    )
 
 
 def memoize(enabled):
     """
     Indicates whether computed values should be cached in memory.
+    Overrides the value of `core__memoize_by_default` when set.
 
     Parameters
     ----------
@@ -109,7 +117,9 @@ def memoize(enabled):
     if not isinstance(enabled, bool):
         raise ValueError(f"Argument must be a boolean; got {enabled!r}")
 
-    return provider_wrapper(AttrUpdateProvider, "_can_memoize", enabled)
+    return decorator_updating_accumulator(
+        lambda acc: acc.update_attr("should_memoize", enabled, "@memoize")
+    )
 
 
 def changes_per_run(enabled=None):
@@ -179,19 +189,19 @@ def changes_per_run(enabled=None):
             return summarize(current_data)
     """
 
-    DEFAULT_VALUE = True
     if callable(enabled):
-        func_or_provider = enabled
-        wrapper = provider_wrapper(AttrUpdateProvider, "changes_per_run", DEFAULT_VALUE)
-        return wrapper(func_or_provider)
+        func = enabled
+        return changes_per_run()(func)
 
     if enabled is None:
-        enabled = DEFAULT_VALUE
+        enabled = True
 
     if not isinstance(enabled, bool):
         raise ValueError(f"Argument must be a boolean; got {enabled!r}")
 
-    return provider_wrapper(AttrUpdateProvider, "changes_per_run", enabled)
+    return decorator_updating_accumulator(
+        lambda acc: acc.wrap_provider(AttrUpdateProvider, "changes_per_run", enabled)
+    )
 
 
 def output(name):
@@ -213,7 +223,9 @@ def output(name):
         A decorator which can be applied to an entity function.
     """
 
-    return provider_wrapper(RenamingProvider, name)
+    return decorator_updating_accumulator(
+        lambda acc: acc.wrap_provider(RenamingProvider, name)
+    )
 
 
 def outputs(*names):
@@ -241,7 +253,9 @@ def outputs(*names):
         A decorator which can be applied to an entity function.
     """
 
-    return provider_wrapper(NameSplittingProvider, names)
+    return decorator_updating_accumulator(
+        lambda acc: acc.wrap_provider(NameSplittingProvider, names)
+    )
 
 
 def docs(*docs):
@@ -264,12 +278,9 @@ def docs(*docs):
         A decorator which can be applied to an entity function.
     """
 
-    return provider_wrapper(AttrUpdateProvider, "docs", docs)
-
-
-# TODO I'd like to put a @protocols decorator here that exposes the
-# MultiProtocolUpdateProvider class, but that would collide with the
-# protocols.py module.  Let's do this in a later PR.
+    return decorator_updating_accumulator(
+        lambda acc: acc.update_attr("docs", docs, "@docs")
+    )
 
 
 def gather(over, also=None, into="gather_df"):
@@ -331,8 +342,14 @@ def gather(over, also=None, into="gather_df"):
     """
     over = interpret.str_or_seq_as_list(over)
     also = interpret.str_or_seq_or_none_as_list(also)
-    return provider_wrapper(
-        GatherProvider, primary_names=over, secondary_names=also, gathered_dep_name=into
+
+    return decorator_updating_accumulator(
+        lambda acc: acc.wrap_provider(
+            GatherProvider,
+            primary_names=over,
+            secondary_names=also,
+            gathered_dep_name=into,
+        )
     )
 
 
@@ -362,20 +379,69 @@ def pyplot(name=None, savefig_kwargs=None):
         A decorator which can be applied to an entity function.
     """
 
-    DEFAULT_NAME = "pyplot"
     if callable(name):
-        func_or_provider = name
-        wrapper = provider_wrapper(PyplotProvider, DEFAULT_NAME)
-        return wrapper(func_or_provider)
+        func = name
+        return pyplot()(func)
 
     if name is None:
-        name = DEFAULT_NAME
-    return provider_wrapper(PyplotProvider, name, savefig_kwargs)
+        name = "pyplot"
+
+    return decorator_updating_accumulator(
+        lambda acc: acc.wrap_provider(PyplotProvider, name, savefig_kwargs)
+    )
+
+
+def accepts(**descriptors_by_arg_name):
+    """
+    Indicates that some of the decorated function's arguments should be supplied
+    according to certain descriptors. Each keyword argument to this decorator
+    corresponds to an argument of the decorated function -- that argument's value
+    will correspond to the descriptor passed to the decorator.
+
+    For example:
+
+        @builder
+        @accepts(pair="x, y")
+        def f(pair, z):
+
+    `f` will be called with two arguments: the first (`pair`) will be a tuple
+    containing the values of entities `x` and `y`, and the second (`z`) will just be
+    the value of entity `z`.
+
+    This decorator is currently experimental and does not have any additional
+    user-facing documentation. It may change in non-backwards-compatible ways.
+    """
+
+    outer_dnodes_by_inner = {
+        entity_dnode_from_descriptor(arg_name): dnode_from_descriptor(descriptor)
+        for arg_name, descriptor in descriptors_by_arg_name.items()
+    }
+
+    return decorator_updating_accumulator(
+        lambda acc: acc.wrap_provider(
+            ArgDescriptorSubstitutionProvider, outer_dnodes_by_inner
+        )
+    )
+
+
+def returns(out_descriptor):
+    """
+    Indicates that the decorated function returns a value corresponding to the provided
+    descriptor.
+
+    This decorator is currently experimental and does not have any additional
+    user-facing documentation. It may change in non-backwards-compatible ways.
+    """
+
+    out_dnode = dnode_from_descriptor(out_descriptor)
+    return decorator_updating_accumulator(
+        lambda acc: acc.wrap_provider(NewOutputDescriptorProvider, out_dnode)
+    )
 
 
 immediate = persist(False)
 immediate.__doc__ = """
-Guarantees that an entity can computed during bootstrap resolution.
+Guarantees that an entity can be computed during bootstrap resolution.
 
 Currently ``@immediate`` is equivalent to ``@persist(False)``.
 """
