@@ -6,6 +6,7 @@ construction and execution APIs (respectively).
 import os
 import shutil
 import warnings
+from collections import defaultdict
 from pathlib import Path, PosixPath
 from importlib import reload
 from textwrap import dedent
@@ -16,23 +17,30 @@ import pandas as pd
 
 # A bit annoying that we have to rename this when we import it.
 from . import protocols as protos
-from .cache import LocalStore, GcsCloudStore, PersistentCache
-from .datatypes import CaseKey, VersioningPolicy
+from .cache_api import Cache
+from .datatypes import (
+    CaseKey,
+    EntityDefinition,
+    VersioningPolicy,
+    ResultGroup,
+)
 from .exception import (
     UndefinedEntityError,
     AlreadyDefinedEntityError,
     IncompatibleEntityError,
+    UnsetEntityError,
+    AttributeValidationError,
 )
+from .executor import Executor
+from .persistence import LocalStore, GcsCloudStore, PersistentCache
 from .provider import (
     ValueProvider,
-    multi_index_from_case_keys,
-    as_provider,
-    provider_wrapper,
+    HashableWrapper,
     AttrUpdateProvider,
 )
-from .deriver import EntityDeriver
-from .descriptors import DescriptorNode
-from . import decorators
+from .deriver import EntityDeriver, entity_is_internal
+from .descriptors.parsing import entity_dnode_from_descriptor
+from . import decorators, decoration
 from .util import (
     group_pairs,
     check_exactly_one_present,
@@ -42,11 +50,10 @@ from .util import (
     oneline,
 )
 
-import logging
-
-logger = logging.getLogger(__name__)
-
 DEFAULT_PROTOCOL = protos.CombinedProtocol(
+    # Every GeoPandas DataFrame is also a Pandas DataFrame, so we need to do the
+    # GeoPandas check first.
+    protos.GeoPandasProtocol(),
     protos.ParquetDataFrameProtocol(),
     protos.ImageProtocol(),
     protos.NumPyProtocol(),
@@ -73,7 +80,8 @@ class FlowState(pyrs.PClass):
     """
 
     providers_by_name = pyrs.field(initial=pyrs.pmap())
-    default_provider_names = pyrs.field(initial=pyrs.pset())
+    entity_defs_by_name = pyrs.field(initial=pyrs.pmap())
+    default_entity_names = pyrs.field(initial=pyrs.pset())
     last_added_case_key = pyrs.field(initial=None)
 
     # These are used to keep track of whether a flow state is safe to reload.
@@ -94,6 +102,47 @@ class FlowState(pyrs.PClass):
     def touch(self):
         return self.set(is_blessed=False)
 
+    def define_entity(self, entity_def):
+        if entity_def.name in self.entity_defs_by_name:
+            raise AlreadyDefinedEntityError(entity_def.name)
+
+        if entity_is_internal(entity_def.name):
+            if entity_def.can_persist:
+                message = f"""
+                Attempted to set @persist to True for Bionic internal
+                entity {entity_def.name!r}.
+                Disable persistence for the decorators by applying
+                `@persist(False)` or `@immediate` to the corresponding
+                function or passing `persist=False` when you
+                `declare` / `assign` the entity values.
+                """
+                raise AttributeValidationError(oneline(message))
+            if entity_def.optional_should_memoize is False:
+                message = f"""
+                Attempted to set @memoize to False for Bionic internal
+                entity {entity_def.name!r}.
+                Enable memoization by removing `@memoize(False)` from the
+                corresponding function.
+                """
+                raise AttributeValidationError(oneline(message))
+
+        return self._set_entity_def(entity_def).touch()
+
+    def delete_entity_defs(self, names):
+        state = self
+
+        for name in names:
+            if name not in self.entity_defs_by_name:
+                continue
+            state = state._erase_entity_def(name).touch()
+
+        return state.touch()
+
+    def get_entity_def(self, name):
+        if name not in self.entity_defs_by_name:
+            raise UndefinedEntityError.for_name(name)
+        return self.entity_defs_by_name[name]
+
     def get_provider(self, name):
         if name not in self.providers_by_name:
             raise UndefinedEntityError.for_name(name)
@@ -102,49 +151,118 @@ class FlowState(pyrs.PClass):
     def has_provider(self, name):
         return name in self.providers_by_name
 
-    def create_provider(self, name, protocol, doc):
+    def create_provider(self, name):
         if name in self.providers_by_name:
             raise AlreadyDefinedEntityError.for_name(name)
 
-        provider = ValueProvider(name, protocol, doc)
+        provider = ValueProvider([name])
         return self._set_provider(provider).touch()
 
     def install_provider(self, provider):
-        for name in provider.attrs.names:
+        for name in provider.entity_names:
             if name in self.providers_by_name:
                 raise AlreadyDefinedEntityError.for_name(name)
 
         return self._set_provider(provider).touch()
 
-    def add_case(self, name, case_key, value):
-        provider = self.get_provider(name).copy_if_mutable()
+    def group_names_together(self, names):
+        providers = [self.get_provider(name) for name in names]
+        providers_set = set(providers)
+
+        # If there's only one provider, then the grouping is already done.
+        if len(providers_set) == 1:
+            provider = providers[0]
+            if set(provider.entity_names) != set(names):
+                if len(names) == 1:
+                    message = f"""
+                    Can't assign to individual entity {names[0]!r}:
+                    it's already defined jointly with entities
+                    {provider.entity_names!r},
+                    and all subsequent assignments must be to that same group
+                    """
+                else:
+                    message = f"""
+                    Can't jointly assign to entities {names!r}:
+                    they're already defined jointly with entities
+                    {provider.entity_names!r},
+                    and all subsequent assignments must be to that same group
+                    """
+                    raise IncompatibleEntityError(oneline(message))
+
+            return self
+
+        # If each name has a different provider, we'll try to combine into a single
+        # provider.
+        elif len(providers_set) == len(names):
+            for provider in providers:
+                if len(provider.entity_names) != 1:
+                    message = f"""
+                    Can't jointly assign to entities {names!r}:
+                    they overlap with existing grouping {provider.entity_names!r}
+                    """
+                    raise IncompatibleEntityError(oneline(message))
+
+                if not isinstance(provider, ValueProvider):
+                    message = f"""
+                    Can't jointly assign to entities {names!r}:
+                    entity {provider.entity_names[0]!r} is already defined as a
+                    function
+                    """
+                    raise IncompatibleEntityError(oneline(message))
+
+                if provider.has_any_cases():
+                    message = f"""
+                    Can't jointly assign to entities {names!r}:
+                    entity {provider.entity_names[0]!r} already has individual
+                    values assigned to it
+                    """
+                    raise IncompatibleEntityError(oneline(message))
+
+            provider = ValueProvider.from_single_value_providers(providers)
+            return self._set_provider(provider).touch()
+
+        else:
+            groupings = [provider.entity_names for provider in providers]
+            message = f"""
+            Can't jointly assign to entities {names!r}:
+            they're already in separate groupings {", ".join(repr(groupings))}
+            """
+            raise IncompatibleEntityError(oneline(message))
+
+    def add_case(self, case_key, names, values):
+        (provider,) = set(self.get_provider(name) for name in names)
         if not isinstance(provider, ValueProvider):
-            raise ValueError(f"Can't add case to function entity {name!r}")
-        provider.check_can_add_case(case_key, value)
-        provider.add_case(case_key, value)
+            raise IncompatibleEntityError(
+                f"Can't add case to function entities {names!r}"
+            )
+
+        # We should already have called group_names_together(), so we know the names
+        # argument should match the names of the provider. However, the names may not
+        # be in the expected order. If not, we'll reorder the values to correspond to
+        # the expected order.
+        provider_names = provider.entity_names
+        if names != provider_names:
+            values = [values[provider_names.index(name)] for name in names]
+
+        tokens = []
+        for name, value in zip(names, values):
+            protocol = self.get_entity_def(name).protocol
+            protocol.validate(value)
+            tokens.append(protocol.tokenize(value))
+        provider = provider.add_case(case_key, values, tokens)
 
         return self._set_provider(provider).touch()
 
     def clear_providers(self, names):
         state = self
 
-        # Remember the original provider attributes per entity.
-        attrs_by_entity_name = {}
-        for name in names:
-            if not state.has_provider(name):
-                continue
-            provider = self.get_provider(name)
-            attrs_by_entity_name[name] = provider.attrs
-
         # Delete the providers (or fail if not possible).
         state = state.delete_providers(names)
 
         # Recreate an empty version of each provider.
-        for name, attrs in attrs_by_entity_name.items():
-            name_ix = attrs.names.index(name)
-            protocol = attrs.protocols[name_ix]
-            doc = attrs.docs[name_ix]
-            state = state.create_provider(name, protocol, doc)
+        entity_names = [name for name in names if name in state.entity_defs_by_name]
+        for name in entity_names:
+            state = state.create_provider(name)
 
         return state.touch()
 
@@ -159,7 +277,7 @@ class FlowState(pyrs.PClass):
                 continue
             provider = state.get_provider(name)
 
-            joint_names = provider.get_joint_names()
+            joint_names = provider.entity_names
             for related_name in joint_names:
                 if related_name not in names:
                     raise IncompatibleEntityError(
@@ -175,35 +293,48 @@ class FlowState(pyrs.PClass):
 
         return state.touch()
 
-    def mark_all_providers_default(self):
+    def mark_all_entities_default(self):
+        new_default_names = pyrs.pset(self.entity_defs_by_name.keys())
+        return self.set(default_entity_names=new_default_names).touch()
+
+    def _erase_entity_def(self, name):
+        assert name not in self.providers_by_name
+
         state = self
-
-        new_default_names = pyrs.pset(state.providers_by_name.keys())
-        state = state.set(default_provider_names=new_default_names)
-
-        return state.touch()
+        state = state.set(entity_defs_by_name=state.entity_defs_by_name.remove(name),)
+        return state
 
     def _erase_provider(self, name):
         state = self
         if name in state.providers_by_name:
-            state = state.set(providers_by_name=state.providers_by_name.remove(name))
-        if name in state.default_provider_names:
+            state = state.set(providers_by_name=state.providers_by_name.remove(name),)
+        if name in state.default_entity_names:
             state = state.set(
-                default_provider_names=(state.default_provider_names.remove(name))
+                default_entity_names=(state.default_entity_names.remove(name))
             )
         return state
 
+    def _set_entity_def(self, entity_def):
+        return self.set(
+            entity_defs_by_name=self.entity_defs_by_name.set(
+                entity_def.name, entity_def
+            ),
+        )
+
     def _set_provider(self, provider):
+        for name in provider.entity_names:
+            assert name in self.entity_defs_by_name
+
         state = self
-        for name in provider.attrs.names:
+        for name in provider.entity_names:
             state = state._erase_provider(name)
             state = state.set(
-                providers_by_name=state.providers_by_name.set(name, provider)
+                providers_by_name=state.providers_by_name.set(name, provider),
             )
         return state
 
 
-class FlowBuilder(object):
+class FlowBuilder:
     """
     A mutable builder for Flows.
 
@@ -250,7 +381,7 @@ class FlowBuilder(object):
         self._state = state.touch()
         return flow
 
-    def declare(self, name, protocol=None, doc=None, docstring=None):
+    def declare(self, name, protocol=None, doc=None, docstring=None, persist=True):
         """
         Creates a new entity but does not assign it a value.
 
@@ -266,6 +397,14 @@ class FlowBuilder(object):
             protocol.
         doc: String, optional
             Description of the new entity.
+        persist: Boolean, optional (default: True)
+            Whether this entity's values should be cached persistently.
+            The only reason to set this to False is if an internal Bionic
+            entity depends on it; in this case, persistence is impossible
+            because Bionic's cache won't be constructed by the time this
+            entity is calculated. The downside of setting this to False is
+            that it won't be possible to retrieve a serialized file for
+            this entity using the mode argument to Flow.get.
         """
 
         if protocol is None:
@@ -279,10 +418,25 @@ class FlowBuilder(object):
             )
             doc = docstring
 
-        self._state = self._state.create_provider(name, protocol, doc)
+        self._state = self._state.define_entity(
+            EntityDefinition(
+                name=name,
+                protocol=protocol,
+                doc=doc,
+                can_persist=persist,
+                optional_should_memoize=True,
+            )
+        ).create_provider(name)
 
     def assign(
-        self, name, value=None, values=None, protocol=None, doc=None, docstring=None
+        self,
+        name,
+        value=None,
+        values=None,
+        protocol=None,
+        doc=None,
+        docstring=None,
+        persist=True,
     ):
         """
         Creates a new entity and assigns it a value.
@@ -304,6 +458,14 @@ class FlowBuilder(object):
             protocol.
         doc: String, optional
             Description of the new entity.
+        persist: Boolean, optional (default: True)
+            Whether this entity's values should be cached persistently.
+            The only reason to set this to False is if an internal Bionic
+            entity depends on it; in this case, persistence is impossible
+            because Bionic's cache won't be constructed by the time this
+            entity is calculated. The downside of setting this to False is
+            that it won't be possible to retrieve a serialized file for
+            this entity using the mode argument to Flow.get.
         """
 
         check_at_most_one_present(value=value, values=values)
@@ -321,15 +483,25 @@ class FlowBuilder(object):
             )
             doc = docstring
 
+        # TODO We can remove this validation, since it also happens in add_case below.
         for value in values:
             protocol.validate(value)
 
         state = self._state
 
-        state = state.create_provider(name, protocol, doc)
+        state = state.define_entity(
+            EntityDefinition(
+                name=name,
+                protocol=protocol,
+                doc=doc,
+                can_persist=persist,
+                optional_should_memoize=True,
+            )
+        )
+        state = state.create_provider(name)
         for value in values:
-            case_key = CaseKey([(name, value, protocol.tokenize(value))])
-            state = state.add_case(name, case_key, value)
+            case_key = CaseKey([(name, protocol.tokenize(value))])
+            state = state.add_case(case_key, [name], [value])
 
         self._state = state
 
@@ -359,16 +531,15 @@ class FlowBuilder(object):
         state = self._state
 
         state = state.clear_providers([name])
-        provider = state.get_provider(name)
-        # This provider must have a single name and single protocol; otherwise
-        # we wouldn't have been able to clear it.
-        (protocol,) = provider.attrs.protocols
 
-        protocol.validate(value)
+        protocol = state.get_entity_def(name).protocol
 
         for value in values:
-            case_key = CaseKey([(name, value, protocol.tokenize(value))])
-            state = state.add_case(name, case_key, value)
+            # TODO We can remove this validation, since it also happens in add_case
+            # below.
+            protocol.validate(value)
+            case_key = CaseKey([(name, protocol.tokenize(value))])
+            state = state.add_case(case_key, [name], [value])
 
         self._state = state
 
@@ -447,27 +618,24 @@ class FlowBuilder(object):
 
         state = self._state
 
-        case_nvt_tuples = []
+        names = [name for name, value in name_value_pairs]
+        state = state.group_names_together(names)
+
+        values = []
+        name_token_pairs = []
         for name, value in name_value_pairs:
-            provider = state.get_provider(name)
-            if len(provider.attrs.protocols) > 1:
-                raise IncompatibleEntityError(
-                    oneline(
-                        f"""
-                    Can't add case for entity co-generated with other
-                    entities {provider.attrs.names!r}"""
-                    )
-                )
-            (protocol,) = provider.attrs.protocols
+            protocol = state.get_entity_def(name).protocol
+            # TODO Both the validation and tokenization are also happening in
+            # state.add_case below; maybe we can remove some duplicate work.
             protocol.validate(value)
             token = protocol.tokenize(value)
 
-            case_nvt_tuples.append((name, value, token))
+            values.append(value)
+            name_token_pairs.append((name, token))
 
-        case_key = CaseKey(case_nvt_tuples)
+        case_key = CaseKey(name_token_pairs)
 
-        for name, value, _ in case_nvt_tuples:
-            state = state.add_case(name, case_key, value)
+        state = state.add_case(case_key, names, values)
 
         case = FlowCase(self, case_key)
         state = state.set(last_added_case_key=case.key)
@@ -507,7 +675,7 @@ class FlowBuilder(object):
             The entities to be deleted.
         """
 
-        self._state = self._state.delete_providers(names)
+        self._state = self._state.delete_providers(names).delete_entity_defs(names)
 
     def merge(self, flow, keep="error", allow_name_match=False):
         """
@@ -560,9 +728,11 @@ class FlowBuilder(object):
             if (
                 old_name_provider is not None
                 and isinstance(old_name_provider, ValueProvider)
-                and len(old_name_provider._values_by_case_key.keys()) == 1
+                and len(old_name_provider._value_tuples_by_case_key.keys()) == 1
             ):
-                (old_flow_name,) = old_name_provider._values_by_case_key.values()
+                (
+                    (old_flow_name,),
+                ) = old_name_provider._value_tuples_by_case_key.values()
                 new_flow_name = flow.name
                 if old_flow_name == new_flow_name and not allow_name_match:
                     raise ValueError(
@@ -609,7 +779,7 @@ class FlowBuilder(object):
                 conflict.resolve("new", "conflicting definition is default")
                 continue
 
-            if conflict.old_protocol is conflict.new_protocol:
+            if conflict.old_entity_def.protocol is conflict.new_entity_def.protocol:
                 if conflict.new_is_only_declaration:
                     conflict.resolve(
                         "old",
@@ -661,7 +831,7 @@ class FlowBuilder(object):
             ("new", new_state),
         ]:
             for provider in state.providers_by_name.values():
-                names = provider.get_joint_names()
+                names = provider.entity_names
                 if len(names) == 1:
                     continue
 
@@ -707,6 +877,7 @@ class FlowBuilder(object):
         names_to_delete = [conflict.name for conflict in conflicts_keeping_new]
         try:
             cur_state = cur_state.delete_providers(names_to_delete)
+            cur_state = cur_state.delete_entity_defs(names_to_delete)
         except IncompatibleEntityError as e:
             raise IncompatibleEntityError("Merge failure: " + e)
 
@@ -714,7 +885,7 @@ class FlowBuilder(object):
         # may have multiple names and hence multiple conflicts, but should be
         # installed exactly once.
         providers_to_install = {
-            tuple(conflict.new_provider.attrs.names): conflict.new_provider
+            tuple(conflict.new_provider.entity_names): conflict.new_provider
             for conflict in conflicts_keeping_new
         }.values()
         for provider in providers_to_install:
@@ -725,14 +896,15 @@ class FlowBuilder(object):
                 not isinstance(provider, ValueProvider)
                 and provider.attrs.orig_flow_name is None
             ):
-                provider = provider_wrapper(
-                    AttrUpdateProvider, "orig_flow_name", new_flow_name
-                )(provider)
+                provider = AttrUpdateProvider(provider, "orig_flow_name", new_flow_name)
+            for name in provider.entity_names:
+                new_entity_def = new_state.get_entity_def(name)
+                cur_state = cur_state.define_entity(new_entity_def)
             cur_state = cur_state.install_provider(provider)
 
         self._state = cur_state
 
-    def __call__(self, func_or_provider):
+    def __call__(self, func):
         """
         Defines an entity by providing a function that derives its value from
         other entities.
@@ -747,80 +919,83 @@ class FlowBuilder(object):
         Parameters
         ----------
 
-        func_or_provider: Function or entity
+        func: Function
             A Python function, optionally decorated with one or more Bionic
             entity decorators.
         """
 
-        provider = as_provider(func_or_provider)
-        if provider.attrs.protocols is None:
-            provider = DEFAULT_PROTOCOL(provider)
-        if provider.attrs.docs is None:
-            docs = [None] * len(provider.attrs.names)
-            provider = decorators.docs(*docs)(provider)
-        if provider.attrs._can_persist is None:
-            provider = decorators.persist(True)(provider)
-        if provider.attrs._can_memoize is None:
-            provider = decorators.memoize(True)(provider)
+        decoration.init_accumulator_if_not_set_on_func(func)
+        acc = decoration.pop_accumulator_from_func(func)
+
+        provider = acc.provider
+
+        # TODO Data like `changes_per_run` and `version` should probably also be stored
+        # in the DecorationAccumulator and then attached to some kind of
+        # FunctionAttributes object.
         if provider.attrs.changes_per_run is None:
-            provider = decorators.changes_per_run(False)(provider)
+            provider = AttrUpdateProvider(provider, "changes_per_run", False)
 
-        if not (provider.attrs.should_persist() or provider.attrs.should_memoize()):
-            raise ValueError(
-                oneline(
-                    f"""
-                Attempted to set both persist and memoize to False.
-                At least one form of storage must be enabled for entities:
-                {func_or_provider.attrs.names!r}"""
-                )
-            )
-        if len(provider.attrs.protocols) != len(provider.attrs.names):
-            names = provider.attrs.names
-            protocols = provider.attrs.protocols
-            raise ValueError(
-                oneline(
-                    f"""
-                Number of protocols must match the number of names;
-                got {len(names)} names {tuple(names)!r} and
-                {len(protocols)} protocols {tuple(protocols)!r}"""
-                )
-            )
-        if len(provider.attrs.docs) != len(provider.attrs.names):
-            names = provider.attrs.names
-            docs = provider.attrs.docs
-            message = oneline(
-                f"""
-                Number of docs must match the number of names;
-                got {len(names)} names {tuple(names)!r} and
-                {len(docs)} docs {tuple(docs)!r}"""
-            )
-            if len(provider.attrs.docs) == 1:
-                prefix = oneline(
-                    """
-                    Using a single doc string for a multi-output function is
-                    deprecated and will become an error condition in a future
-                    release; use the ``@docs`` decorator to specify
-                    one doc string for each entity.  Details:"""
-                )
-                warnings.warn(prefix + "\n" + message)
+        names = provider.entity_names
 
-                docs = [docs[0]] * len(names)
-                provider = AttrUpdateProvider(
-                    provider, "docs", docs, allow_override=True
-                )
-            else:
-                raise ValueError(message)
+        protocol = acc.protocol
+        if protocol is None:
+            protocol = DEFAULT_PROTOCOL
+
+        docs = acc.docs
+        orig_docstring = provider.get_source_func().__doc__
+        if docs is None:
+            if orig_docstring is not None and len(names) > 1:
+                preamble = """
+                Using a single doc string for a multi-output function is
+                deprecated and will become an error condition in a future
+                release; use the ``@docs`` decorator to specify
+                one doc string for each entity.  Details:
+                """
+                details = f"""
+                got {len(names)} names {tuple(names)!r} and
+                just one docstring {orig_docstring}
+                """
+                warnings.warn(oneline(preamble) + "\n" + oneline(details))
+
+            docs = [orig_docstring for name in names]
+        if len(docs) != len(names):
+            message = f"""
+            Number of docs must match the number of names;
+            got {len(names)} names {tuple(names)!r} and
+            {len(docs)} docs {tuple(docs)!r}"""
+            raise ValueError(oneline(message))
+
+        can_persist = acc.can_persist
+        if can_persist is None:
+            can_persist = True
 
         state = self._state
 
-        state = state.delete_providers(provider.attrs.names)
+        # Delete the original definitions.
+        # TODO Deleting the original entity definition is a little counterintuitive. It
+        # would probably make more sense to keep the original definition, although
+        # that would be a breaking change at this point. We need to think through the
+        # relationship between APIs for fixed and derived entities.
+        state = state.delete_providers(provider.entity_names)
+        state = state.delete_entity_defs(provider.entity_names)
+
+        # Create the new definitions.
+        for name, doc in zip(names, docs):
+            entity_def = EntityDefinition(
+                name=name,
+                protocol=protocol,
+                doc=doc,
+                can_persist=can_persist,
+                optional_should_memoize=acc.should_memoize,
+            )
+            state = state.define_entity(entity_def)
         state = state.install_provider(provider)
 
         self._state = state
 
         return provider.get_source_func()
 
-    def derive(self, func_or_provider):
+    def derive(self, func):
         """
         (Deprecated) An alias for ``__call__``; use that instead.
         """
@@ -830,7 +1005,7 @@ class FlowBuilder(object):
             "decorator) instead."
         )
 
-        return self.derive(func_or_provider)
+        return self.derive(func)
 
     # --- Private helpers.
 
@@ -846,7 +1021,7 @@ class FlowBuilder(object):
         self.set("core__flow_name", name)
 
     def _set_for_case_key(self, case_key, name, value):
-        self._state = self._state.add_case(name, case_key, value)
+        self._state = self._state.add_case(case_key, [name], [value])
 
     def _set_for_last_case(self, name, value):
         last_case_key = self._state.last_added_case_key
@@ -856,26 +1031,15 @@ class FlowBuilder(object):
         self._set_for_case_key(last_case_key, name, value)
 
 
-class MergeConflict(object):
+class MergeConflict:
     def __init__(self, old_state, new_state, name):
         self.name = name
         self.old_provider = old_state.providers_by_name.get(name)
         self.new_provider = new_state.providers_by_name.get(name)
-        self.old_is_default = name in old_state.default_provider_names
-        self.new_is_default = name in new_state.default_provider_names
-
-    @property
-    def old_protocol(self):
-        return self._protocol_for_provider(self.old_provider)
-
-    @property
-    def new_protocol(self):
-        return self._protocol_for_provider(self.new_provider)
-
-    def _protocol_for_provider(self, provider):
-        assert provider is not None
-        name_ix = provider.attrs.names.index(self.name)
-        return provider.attrs.protocols[name_ix]
+        self.old_is_default = name in old_state.default_entity_names
+        self.new_is_default = name in new_state.default_entity_names
+        self.old_entity_def = old_state.entity_defs_by_name.get(name)
+        self.new_entity_def = new_state.entity_defs_by_name.get(name)
 
     @property
     def old_is_only_declaration(self):
@@ -893,7 +1057,7 @@ class MergeConflict(object):
         self.reason = reason
 
 
-class FlowCase(object):
+class FlowCase:
     """
     A specific case for which entities can have associated values.
 
@@ -910,7 +1074,7 @@ class FlowCase(object):
         return self
 
 
-class Flow(object):
+class Flow:
     """
     An immutable workflow object.  You can use get() to compute any entity
     in the workflow, or setting() to create a new workflow with modifications.
@@ -933,7 +1097,7 @@ class Flow(object):
         return [
             name
             for name in self._state.providers_by_name.keys()
-            if include_core or not self._deriver.entity_is_internal(name)
+            if include_core or not entity_is_internal(name)
         ]
 
     def entity_protocol(self, name):
@@ -947,7 +1111,7 @@ class Flow(object):
             The name of an entity.
         """
 
-        return self._state.get_provider(name).protocol_for_name(name)
+        return self._state.get_entity_def(name).protocol
 
     def entity_doc(self, name):
         """
@@ -960,7 +1124,7 @@ class Flow(object):
         name: String
             The name of an entity.
         """
-        return self._state.get_provider(name).doc_for_name(name)
+        return self._state.get_entity_def(name).doc
 
     def entity_docstring(self, name):
         """
@@ -1032,8 +1196,15 @@ class Flow(object):
         The value of the entity, or a collection containing its values.
         """
 
-        dnode = DescriptorNode.from_descriptor(name)
-        result_group = self._deriver.derive(dnode)
+        dnode = entity_dnode_from_descriptor(name)
+        orig_result_group = self._deriver.derive(dnode)
+        # Remove all results with missing values.
+        result_group = ResultGroup(
+            results=[
+                result for result in orig_result_group if not result.value_is_missing
+            ],
+            key_space=orig_result_group.key_space,
+        )
 
         if mode is object or mode == "object":
             values = [result.value for result in result_group]
@@ -1071,7 +1242,17 @@ class Flow(object):
 
         if collection is None or collection is object:
             if len(values) == 0:
-                raise ValueError(f"Entity {name!r} has no defined values")
+                # There should always be at least one result, but some of them may have
+                # missing values.
+                assert len(orig_result_group) > 0
+                missing_result = orig_result_group[0]
+                missing_names = missing_result.query.case_key.missing_names
+                message = f"""
+                Entity {name} could not be computed because the following entities are
+                declared but have no values set:
+                {", ".join(repr(name) for name in missing_names)}
+                """
+                raise UnsetEntityError(oneline(message))
             if len(values) > 1:
                 raise ValueError(f"Entity {name!r} has multiple values")
             return values[0]
@@ -1080,10 +1261,42 @@ class Flow(object):
         elif collection is set or collection == "set":
             return set(values)
         elif collection is pd.Series or collection == "series":
-            if len(result_group.key_space) > 0:
-                index = multi_index_from_case_keys(
-                    case_keys=[result.query.case_key for result in result_group],
-                    ordered_key_names=list(result_group.key_space),
+            key_space = list(result_group.key_space)
+            if len(key_space) > 0:
+                ancestor_values_by_case_key_by_name = defaultdict(dict)
+                ancestor_key_space_by_name = {}
+                for ancestor_name in key_space:
+                    ancestor_result_group = self._deriver.derive(
+                        entity_dnode_from_descriptor(ancestor_name)
+                    )
+                    ancestor_key_space_by_name[
+                        ancestor_name
+                    ] = ancestor_result_group.key_space
+                    for ancestor_result in ancestor_result_group:
+                        ancestor_case_key = ancestor_result.query.case_key
+                        ancestor_values_by_case_key_by_name[ancestor_name][
+                            ancestor_case_key
+                        ] = ancestor_result.value
+
+                orig_case_keys = [result.query.case_key for result in result_group]
+                index = pd.MultiIndex.from_tuples(
+                    tuples=[
+                        tuple(
+                            HashableWrapper(
+                                value=ancestor_values_by_case_key_by_name[
+                                    ancestor_name
+                                ][
+                                    orig_case_key.project(
+                                        ancestor_key_space_by_name[ancestor_name]
+                                    )
+                                ],
+                                token=orig_case_key.tokens[ancestor_name],
+                            )
+                            for ancestor_name in key_space
+                        )
+                        for orig_case_key in orig_case_keys
+                    ],
+                    names=list(key_space),
                 )
             else:
                 index = None
@@ -1091,7 +1304,6 @@ class Flow(object):
         else:
             raise ValueError(f"Unrecognized collection type {collection!r}")
 
-    # TODO Maybe this wants to be two different functions?
     def export(self, name, file_path=None, dir_path=None):
         """
         Provides access to the persisted file corresponding to an entity.  Note:
@@ -1119,7 +1331,7 @@ class Flow(object):
             "Flow#export is deprecated; use the mode argument to Flow#get " "instead"
         )
 
-        dnode = DescriptorNode.from_descriptor(name)
+        dnode = entity_dnode_from_descriptor(name)
         result_group = self._deriver.derive(dnode)
         if len(result_group) != 1:
             raise ValueError(
@@ -1384,6 +1596,8 @@ class Flow(object):
 
         self.get = GetMethod()
 
+        self.cache = Cache(self._deriver)
+
         class SettingMethod(ShortcutProxy):
             __doc__ = orig_setting_method.__doc__
 
@@ -1405,7 +1619,7 @@ class Flow(object):
         return Flow._from_state(builder._state)
 
 
-class ShortcutProxy(object):
+class ShortcutProxy:
     """
     Wraps a method on a Flow object to allow it to be called via an alternative
     style.
@@ -1472,10 +1686,11 @@ class ShortcutProxy(object):
 def create_default_flow_state():
     builder = FlowBuilder._with_empty_state()
 
-    builder.declare("core__flow_name")
+    builder.declare("core__flow_name", persist=False)
 
-    builder.assign("core__persistent_cache__global_dir", "bndata")
-    builder.assign("core__versioning_mode", "manual")
+    builder.assign("core__memoize_by_default", True, persist=False)
+    builder.assign("core__persistent_cache__global_dir", "bndata", persist=False)
+    builder.assign("core__versioning_mode", "manual", persist=False)
 
     @builder
     @decorators.immediate
@@ -1509,8 +1724,8 @@ def create_default_flow_state():
     ):
         return os.path.join(core__persistent_cache__global_dir, core__flow_name)
 
-    builder.assign("core__persistent_cache__gcs__bucket_name", None)
-    builder.assign("core__persistent_cache__gcs__enabled", False)
+    builder.assign("core__persistent_cache__gcs__bucket_name", None, persist=False)
+    builder.assign("core__persistent_cache__gcs__enabled", False, persist=False)
 
     @builder
     @decorators.immediate
@@ -1566,4 +1781,17 @@ def create_default_flow_state():
             cloud_store=core__persistent_cache__cloud_store,
         )
 
-    return builder._state.mark_all_providers_default()
+    builder.assign("core__parallel_execution__enabled", False, persist=False)
+    # The executor uses max available CPUs when set to None.
+    builder.assign("core__parallel_execution__worker_count", None, persist=False)
+
+    @builder
+    @decorators.immediate
+    def core__executor(
+        core__parallel_execution__enabled, core__parallel_execution__worker_count,
+    ):
+        if not core__parallel_execution__enabled:
+            return None
+        return Executor(core__parallel_execution__worker_count)
+
+    return builder._state.mark_all_entities_default()
