@@ -13,10 +13,11 @@
 
 import os
 
-from ..helpers import count_calls
-
 import bionic as bn
 
+from ..conftest import ResettingCounter
+from ..helpers import count_calls
+
 
 CACHE_TEST_DIR = os.path.join(
     os.path.dirname(__file__), "test_persistence_compatibility"
@@ -25,13 +26,10 @@ CACHE_TEST_DIR = os.path.join(
 
 class Harness:
     """
-    Holds a simple Bionic flow alongside the different functions used in it.
-    The functions are exposed separately since the tests that use the flow
-    asserts the number of times the methods were invoked in the flow using
-    the @count_calls annotation which is not exposed by Bionic.
+    Holds a simple Bionic flow with counters to all the functions in it.
     """
 
-    def __init__(self, cache_dir):
+    def __init__(self, cache_dir, make_counter):
         builder = bn.FlowBuilder("test")
 
         builder.set("core__persistent_cache__flow_dir", cache_dir)
@@ -39,29 +37,39 @@ class Harness:
         builder.assign("y", 3)
         builder.assign("z", 4)
 
+        xy_counter = make_counter()
+
         @builder
-        @count_calls
+        @count_calls(xy_counter)
         def xy(x, y):
             return x * y
 
+        yz_counter = make_counter()
+
         @builder
-        @count_calls
+        @count_calls(yz_counter)
         def yz(y, z):
             return y * z
 
+        xy_plus_yz_counter = make_counter()
+
         @builder
-        @count_calls
+        @count_calls(xy_plus_yz_counter)
         def xy_plus_yz(xy, yz):
             return xy + yz
 
         self.flow = builder.build()
-        self.xy = xy
-        self.yz = yz
-        self.xy_plus_yz = xy_plus_yz
+        self.xy_counter = xy_counter
+        self.yz_counter = yz_counter
+        self.xy_plus_yz_counter = xy_plus_yz_counter
 
 
 if __name__ == "__main__":
-    flow = Harness(CACHE_TEST_DIR).flow
+
+    def make_counter():
+        return ResettingCounter()
+
+    flow = Harness(CACHE_TEST_DIR, make_counter).flow
 
     # call methods to write to cache
     flow.get("xy")
